\ Cards library for DX-Forth
\ For ANSI terminal and PC 437 charset (depended words are marked by "#####")
\ Pavel Zampach, 2021

include random
decimal

$da constant GULC			\ Upper left corner #####
$bf constant GURC			\ Upper right corner #####
$c0 constant GLLC			\ Lower left corner #####
$d9 constant GLRC			\ Lower right corner #####
$c4 constant GHL			\ Horizontal line #####
$b3 constant GVL			\ Vertical line #####
$b1 constant GHB			\ Hashed block #####
 52 constant #DECK

variable CARDPTR			\ Pointer to card deck

\ Card: low byte is card value (0..13), high byte is card suite
\ Value: 0 = "zero card" (for base of card foundation), 1 = ace .. 13 = king
\ Suite: 0 = heart, 1 = diamond, 2 = club, 3 = spade
create DECK
1 c, 0 c, 2 c, 0 c, 3 c, 0 c,  4 c, 0 c,  5 c, 0 c,  6 c, 0 c,
7 c, 0 c, 8 c, 0 c, 9 c, 0 c, 10 c, 0 c, 11 c, 0 c, 12 c, 0 c, 13 c, 0 c,
1 c, 1 c, 2 c, 1 c, 3 c, 1 c,  4 c, 1 c,  5 c, 1 c,  6 c, 1 c,
7 c, 1 c, 8 c, 1 c, 9 c, 1 c, 10 c, 1 c, 11 c, 1 c, 12 c, 1 c, 13 c, 1 c,
1 c, 2 c, 2 c, 2 c, 3 c, 2 c,  4 c, 2 c,  5 c, 2 c,  6 c, 2 c,
7 c, 2 c, 8 c, 2 c, 9 c, 2 c, 10 c, 2 c, 11 c, 2 c, 12 c, 2 c, 13 c, 2 c,
1 c, 3 c, 2 c, 3 c, 3 c, 3 c,  4 c, 3 c,  5 c, 3 c,  6 c, 3 c,
7 c, 3 c, 8 c, 3 c, 9 c, 3 c, 10 c, 3 c, 11 c, 3 c, 12 c, 3 c, 13 c, 3 c,

\ n = 0..51
: DECK()      ( n -- adr )
  cells deck +
;

\ Horizontal line from 4 chars [c]
: HLINE       ( c -- )
  dup 2dup emit emit emit emit
;

\ Chars for card value
create CARDCHAR
    bl c, char A c, char 2 c, char 3 c, char 4 c, char 5 c, char 6 c,
char 7 c, char 8 c, char 9 c, char 1 c, char J c, char Q c, char K c,

\ Set cursor at [x y] and leave coords of the next line
: NXTLINE     ( x y -- x y+1 )
  2dup at-xy 1+
;

\ Draw card frame at [x y] filled by char [c]
\ Card is 6 chars wide and 5 lines high
: CARDFRAME   ( x y c -- )
  -rot nxtline
  gulc emit ghl hline gurc emit
  3 0 do
    nxtline
    gvl emit 2 pick hline gvl emit
  loop
  at-xy drop
  gllc emit ghl hline glrc emit
;

\ ---------- Interface ----------

\ Increment byte on [addr]
code C++      ( addr -- )
       H pop
       M inr
         next
end-code

\ Decrement byte on [addr]
code C--     ( addr -- )
       H pop
       M dcr
         next
end-code

\  #####
: INVERSE     ( -- )
  27 emit ." [7m"
;

\  #####
: BLINK       ( -- )
  27 emit ." [5m"
;

\ Shuffle card deck
: SHUFFLE     ( -- )
  randomize
  200 0 do
    #deck random deck() dup @
    #deck random deck() dup @
    -rot ! swap !
  loop
  cardptr off
;

\ Get new card
: GETNEWCARD  ( -- u )
  cardptr @ #deck = if shuffle then
  cardptr @ deck() @
  cardptr c++
;

\ Card [u] => card value [b] (0..13)
: CARDVAL     ( u -- b )
  $0f and
;

\ Card [u] => card suite [b] (0..3)
: SUITE       ( u -- b )
  >< 3 and
;

\ [flag] is TRUE if cards [u1] and [u2] have the same suite color (black/red)
: =COLOR      ( u1 u2 -- flag )
  xor $0200 and 0=
;

\ Print char of card suite [b] #####
: .SUITE      ( b -- )
  dup 1 > if inverse then
  3 + emit normal
;

: EMPTYFRAME  ( x y -- )
  bl cardframe
;

: DISPLAYCARD ( card x y -- )
  2dup emptyframe
  1+ swap 1+ swap at-xy
  dup cardval
  dup cardchar + c@ emit
  10 = if [char] 0 emit else space then
  space
  suite .suite
;

: DISPLAYBACK ( x y -- )
  ghb cardframe
;

: DELETECARD  ( x y -- )
  5 0 do
    nxtline 6 spaces
  loop   2drop
;
