\ Fifteenn solitaite game
\ For terminal with 30 lines
\ Pavel Zampach, 2021

true \ FALSE for debug, TRUE for standalone
decimal
include cards

 15 constant #PILLS			\ no of pills
  4 constant PILLMAX			\ max. card in pill
variable GAMEOVER			\ TRUE if game is over
variable STACKPTR			\ addr of first free place on stack
0 value START				\ pill for move from
0 value TARGET				\ pill for move to

dup [IF]
  include statmem
  $8000 set-limit
  $8000 2048 empty-memory

  #pills pillmax * cells allocate mem-check
    constant CARDARRAY
  #pills cells allocate mem-check
    constant CARDNO
  256 allocate mem-check
    constant STACK
[ELSE]
  create CARDARRAY
    #pills pillmax * cells allot
  create CARDNO
    #pills cells allot
  create STACK
    256 allot
[THEN]

create COORDATA
 8 c,  2 c, 17 c,  2 c, 26 c,  2 c, 35 c,  2 c, 44 c,  2 c,
53 c,  2 c, 62 c,  2 c, 71 c,  2 c, 12 c, 14 c, 21 c, 14 c,
30 c, 14 c, 39 c, 14 c, 48 c, 14 c, 57 c, 14 c, 66 c, 14 c,

\ Coords [x y] of pill [a] (0..14)
: COORDS    ( a -- x y )
  2* coordata + dup c@ swap 1+ c@
;

\ (0..14, 0..3)
: CARDARRAY() ( a b -- addr )
  swap pillmax * + cells cardarray +
;

\ Number of cards in pill [a] (0..14)
: CARDNO()    ( a -- addr )
  cells cardno +
;

\ stack data: high nibble = target, low nibble = start
: STACKTOP     ( -- addr )
  stackptr @ stack +
;

\ Index [a b] of top card in pill [a] (0..14)
: TOPIDX      ( a -- a b )
  dup cardno() @ 1-
;

\ Address of top card in pill [a] (0..14)
: TOPCARD     ( a -- addr )
  topidx cardarray()
;

\ Read key
: INPUT       ( -- c )
  1 28 at-xy key upcase
;

\ Print char of pill [a] (0..14)
: PILLCHAR    ( a -- )
  dup coords swap 1- swap at-xy
  [char] A + emit
;

\ Test char of pill
: PILLOK?     ( c -- flag )
  [char] A [char] O between
;

\ Index [a b] of card (0..14, 0..3) to coords [x y]
: COXY        ( a b -- x y )
   2* swap coords rot +
;

\ Draw card of index [a b]
: DRAWCARD    ( a b -- )
  2dup cardarray() @ -rot
  coxy displaycard
;

\ Wipe card of index [a b]
: WIPECARD    ( a b -- )
  coxy deletecard
;

\ Init screen and data
: INIT        ( -- )
  stackptr off
  gameover off
  shuffle
  page
  #pills 0 do
    i 3 4 between if
      i cardno() off			\ pills 3 and 4 are empty
    else
      pillmax i cardno() !		\ fill pill
      pillmax 0 do
        getnewcard j i cardarray() !
        j i drawcard
      loop
    then
    i pillchar
  loop
  5 28 at-xy ." A-O = pill, [del] = backstep, X = exit"
;

\ Test end of game
: TESTEND     ( -- )
  true
  #pills 0 do
    i cardno() @ 1 3 between if		\ 1..3 cards in pill
      drop false
    break
    i cardno() @ if			\ 4 cards in pill
      pillmax 1 do			\ have cards in pill the same value?
        j 0 cardarray() @ cardval
        j i cardarray() @ cardval = and
      loop
    then
  loop
  gameover !
;

\ Screen on end of game
: ENDGAME    ( -- flag )
  page
  32 10 at-xy
  gameover @ if
    ." Congratulations!"
    35 5 at-xy ." Game over"
    35 15 at-xy stackptr ? ." moves"
  else
    ." Game interrupted"
  then
  34 25 at-xy ." Play again? " y/n not
;

\ Store [start] and [target] to stack
: !STACK      ( -- )
  target 4 lshift start or stacktop c!
  stackptr c++
;

\ Move card from [start] to [target]
: MOVECARD    ( -- )
  target cardno() c++			\ inc caddrno(target)
  start topcard @ target topcard !	\ move card
  start topidx wipecard
  start cardno() c--			\ dec cardno(start)
  start cardno() @ if
    start topidx drawcard
  then
  target topidx drawcard
;

\ Moving card from pill of char [c]
: MOVING      ( c -- )
  [char] A - dup to start
  cardno() @ 0= if end			\ [start] is empty
  blink start pillchar
  input
  normal start pillchar
  dup [char] A - to target
  pillok? not if end

  target cardno() @ dup pillmax <	\ [target] is not full
  swap if				\ [target] is not empty
    start topcard @ cardval		\ and card values are the same
    target topcard @ cardval = and
  then
  target start <> and if		\ and pills are different
    movecard
    !stack
    testend
  then
;

: BACKSTEP    ( -- )
  stackptr @ if
    stackptr c--
    stacktop c@ dup
    4 rshift to start
    15 and to target
    movecard
  then
;

: MAIN        ( -- )
  begin
    init
    begin
      input
      dup pillok? if dup moving then
      dup  8 = over 127 = or if backstep then
      [char] X = gameover @ or
    until
    endgame
  until   page
;

[IF]
  turnkey main fifteen
[THEN]
