\ Freecell solitaire game
\ For terminal with 30 lines
\ Pavel Zampach, 2021

true \ FALSE for debug, TRUE for standalone
decimal
include cards

  8 constant #PILLS			\ end of pills index
 12 constant #OPCELL			\ end of open cell index
 16 constant #FOUND			\ end of foundation index
 13 constant PILLMAX			\ max. card in pill
variable GAMEOVER			\ TRUE if game is over
variable STACKPTR			\ addr of first free place in stack
0 value START				\ place for move from
0 value TARGET				\ place for move to

dup [IF]
  include statmem
  $8000 set-limit
  $8000 2048 empty-memory

  #pills pillmax * cells allocate mem-check
    constant CARDARRAY
  #found cells allocate mem-check
    constant HELP
  256 allocate mem-check
    constant STACK
[ELSE]
  create CARDARRAY
    #pills pillmax * cells allot
  create HELP
    #found cells allot
  create STACK
    256 allot
[THEN]

create COORDATA
 9 c,  0 c, 17 c,  0 c, 25 c,  0 c, 33 c,  0 c,
41 c,  0 c, 49 c,  0 c, 57 c,  0 c, 65 c,  0 c,
 1 c,  2 c,  1 c,  9 c,  1 c, 16 c,  1 c, 23 c,
73 c,  2 c, 73 c,  9 c, 73 c, 16 c, 73 c, 23 c,

\ coords [x y] of place [a] (0..15)
: COORDS    ( a -- x y )
  2* coordata + dup c@ swap 1+ c@
;

\ (0..7, 0..12)
: CARDARRAY() ( a b -- addr )
  swap pillmax * + cells cardarray +
;

\ [a] = number of cards in pill (0..7)
\     = card in open cell (8..11)
\     = card in foundation (12..15)
: HELP()      ( a -- addr )
  cells help +
;
\ Addr of help(start)
: START()     ( -- addr )
  start help()
;

\ Addr of help(target)
: TARGET()    ( -- addr )
  target help()
;

\ Stack data: high nibble = target, low nibble = start
: STACKTOP     ( -- addr )
  stackptr @ stack +
;

\ Index [a b] of top card in pill [a] (0..7)
: TOPIDX      ( a -- a b )
  dup help() @ 1-
;

\ Address of top card in pill [a] (0..7)
: TOPCARD     ( a -- addr )
  topidx cardarray()
;

\ Read key
: INPUT       ( -- c )
  0 29 at-xy key upcase
;

\ Print char of pill [a] (0..15)
: PILLCHAR    ( a -- )
  dup coords swap 1- swap at-xy
  [char] A + emit
;

\ Index [a b] of card (0..8, 0..12) to coords [x y]
: CARDXY        ( a b -- x y )
   2* swap coords rot +
;

\ Draw card of index [a b]
: DRAWCARD    ( a b -- )
  2dup cardarray() @ -rot
  cardxy displaycard
;

\ Wipe card of index [a b]
: WIPECARD    ( a b -- )
  cardxy deletecard
;

\ Draw card in open cell or foundation (8..15)
: DRAWOPFOU   ( a card -- )
  swap coords displaycard
;

\ Wipe card in open cell (8..11)
: WIPEOPCELL  ( a -- )
  coords emptyframe
;

\ Init screen and data
: INIT        ( -- )
  stackptr off
  gameover off
  shuffle
  page
  #pills 0 do
    7 i 4 and 2/ 2/ -			\ i 4 < if 7 else 6 then
    dup i help() ! 			\ init help()
    0 do
      getnewcard j i cardarray() !
      j i drawcard
    loop
    i pillchar
  loop
  #opcell #pills do
    i wipeopcell
    i help() off
    i pillchar
  loop
  #found #opcell do
    i dup #opcell - ><
    dup i help() !			\ init help()
    drawopfou
    i pillchar
  loop
  5 29 at-xy ." A-O = pill, Q = auto/open, [del] = backstep, X = exit"
;

\ Fetch card from pill
: PILL>       ( -- card )
  start topcard @
  start topidx wipecard
  start() c--
  start() @ if
    start topidx drawcard
  then
;

\ Fetch card from open cell
: OPCELL>     ( -- card )
  start() @
  start wipeopcell
  start() off
;

\ Fetch card from foundation
: FOUND>      ( -- card )
  start() @
  start() c--
  start start() @ drawopfou
;

\ Store card to pill
: >PILL       ( card -- )
  target() c++
  target topcard !
  target topidx drawcard
;

\ Store card to open cell or foundation
: >OPFOU      ( card -- )
  dup target() !
  target swap drawopfou
;

create START>
' pill> , ' pill> , ' opcell> , ' found> ,

create >TARGET
' >pill , ' >pill , ' >opfou , ' >opfou ,

\ Move card from [start] to [target]
: MOVECARD    ( -- )
  start 2/ 2/ cells start> + @execute
  target 2/ 2/ cells >target + @execute
;

\ Check [start] value
\ moving isn't possible -> [flag] is TRUE
: STARTCHECK  ( -- flag )
  start() @ 0=
;

\ Get card to move (from pill or open cell]
: MCARD       ( -- card )
  start 8 < if
    start topcard @
  else
    start() @
  then
;

\ Check [target] value
\ moving isn't possible -> [flag] is TRUE
: TARGETCHECK ( -- flag )
  target #pills < if
    mcard target topcard @
    2dup =color
    -rot cardval swap cardval
    1+ <> or
    target() @ pillmax = or
    target() @ and
  end
  target #opcell < if
    target() @
  end
  target #found < if
    mcard target() @ 1+ <>
  end
  target #found = if			\ "Q" key pressed
    #opcell #pills do
      i help() @ not if
        i to target
        unloop false
      end
    loop
  then
  true					\ [target] is out of range
;

\ Test end of game (in all foundations are kings)
: TESTEND     ( -- )
  true
  #found #opcell do
    i help() @ cardval 13 = and
  loop
  gameover !
;

\ Screen on end of game
: ENDGAME    ( -- flag )
  page
  32 10 at-xy
  gameover @ if
    ." Congratulations!"
    35 5 at-xy ." Game over"
    35 15 at-xy stackptr ? ." moves"
  else
    ." Game interrupted"
  then
  34 25 at-xy ." Play again? " y/n not
;

\ Store [start] and [target] to stack
: !STACK      ( -- )
  target 4 lshift start or stacktop c!
  stackptr c++
;

\ Moving card from place of char [c]
: MOVING      ( c -- )
  [char] A - to start
  startcheck if end
  blink start pillchar
  input
  normal start pillchar
  [char] A - to target
  targetcheck if end
  movecard
  !stack
  testend
;

: BACKSTEP    ( -- )
  stackptr @ if
    stackptr c--
    stacktop c@ dup
    4 rshift to start
    15 and to target
    movecard
  then
;

\ Automatic move all available cards to foundation
: AUTOFOU     ( -- )
  begin
    true
    #opcell 0 do
      i to start
      startcheck not if
        #found #opcell do
          i to target
          targetcheck not if
            movecard
            !stack
            testend
            drop false
          then
        loop
      then
    loop
  until
;

: MAIN        ( -- )
  begin
    init
    begin
      input
      dup [char] A [char] P between if dup moving then
      dup 8 = over 127 = or if backstep then
      dup [char] Q = if autofou then
      [char] X = gameover @ or
    until
    endgame
  until   page
;

[IF]
  turnkey main freecell
[THEN]
